#include "BnL.h"



BlockNestedLoops::BlockNestedLoops()
{
}


BlockNestedLoops::BlockNestedLoops(char *infile)
{
	fstream fin;
	fin.open(infile, ios::in);
	fin >> numDimensions; // read dimensionality

	// read dimension sizes
	dimSizes = new int[numDimensions];
	for (int d = 0; d < numDimensions; d++)
	{ 
		fin >> dimSizes[d];
		dimSizes[d]++;

		#ifdef DEBUG
		cout << dimSizes[d] << " ";
		#endif
	}
	#ifdef DEBUG
	cout << endl;
	#endif

	// read number of points
	fin >> numPoints; 

	// read points
	int* point;
	for (int p = 0; p < numPoints; p++)
	{
		point = new int[numDimensions+1];

		#ifdef DEBUG
		cout << "point " << p << ": ";
		#endif

		point[0] = p;
		for (int d = 1; d < numDimensions+1; d++)

		{
			fin >> point[d];

			#ifdef DEBUG
			cout << point[d] << " ";
			#endif
		}
		points.push_back(point);

		#ifdef DEBUG
		cout << endl;
		#endif
	}
	fin.close();

	#ifdef DEBUG
	cout << "numDimensions=" << numDimensions << " numPoints=" << numPoints << endl;
	#endif

}


bool BlockNestedLoops::dominates(int *a, int *b)
{
	bool check = false;

	#ifdef DEBUG
	cout << "does ";
	for (int d = 0; d < numDimensions+1; d++)
		cout << a[d] << " ";
	cout << " dominates ";
	for (int d = 0; d < numDimensions+1; d++)
		cout << b[d] << " ";
	#endif

	for (int d = 1; d < numDimensions+1; d++)
	{
		if (a[d] > b[d])
		{
			#ifdef DEBUG
			cout << " NO";
			cout <<  endl;
			#endif

			return false;
		}
		else if (a[d] < b[d])
			check = true;
	}

	#ifdef DEBUG
	if (check)
		cout << " YES";
	else
		cout << " NO";
	cout <<  endl;
	#endif
	
	return check;
}


bool BlockNestedLoops::isDominatedBy(int *a, int *b)
{
	bool check = false;
	for (int d = 1; d < numDimensions+1; d++)
	{
		if (a[d] < b[d])
			return false;
		else if (a[d] > b[d])
			check = true;
	}
	
	return check;
}


vector<int*> BlockNestedLoops::getSkyline(int wsize)
{
	#ifdef PRINT_RESULTS
	cout << endl << "Skyline block nested loops method" << endl;
	#endif

	vector<int*>::iterator pIter;
	vector<int*> skyline;
	vector<int*> window;
	vector<int*>::iterator wIter;
	vector<int*> tempFile;
	int countIn = 0, countOut = 0;
	int prevCountIn = 0;
	map<int, int> timestamp;
	map<int, int>::iterator sIter;

	pIter = points.begin();
	while (pIter != points.end())
	{
		#ifdef DEBUG
		cout << "examine point: ";
		for (int d = 1; d < numDimensions+1; d++)
			cout << (*pIter)[d] << " ";
		cout << " time stamp: " << countOut << " countIn: " << countIn << endl;
		#endif

		wIter = window.begin();
		while (wIter != window.end())
		{
			sIter = timestamp.find((*wIter)[0]);
			if (sIter->second == countIn)
			{
				#ifdef DEBUG
				cout << "case 1\n";
				cout << "output point: ";
				for (int d = 1; d < numDimensions+1; d++)
					cout << (*wIter)[d] << " ";
				cout << " time stamp: " <<  countIn << endl;
				#endif

				skyline.push_back((*wIter));
				window.erase(wIter);
			}

			// Time stamp correction.
			else if ((sIter->second >= points.size()) && (countIn == 0))
			{
				sIter->second = prevCountIn-1;
				++wIter;
			}
			else
				++wIter;
		}
		
		sIter = timestamp.find((*pIter)[0]);
		if (sIter == timestamp.end())
			timestamp.insert(make_pair((*pIter)[0], countOut));
		else
			sIter->second = countOut;
		countIn++;
		bool check = true;

		wIter = window.begin();
		while (wIter != window.end())
		{
			if (dominates((*pIter), (*wIter)))
			{
				#ifdef DEBUG
				cout << "case 2\n";
				#endif

				window.erase(wIter);

				#ifdef DEBUG
				cout << "erase point: ";
				for (int d = 1; d < numDimensions+1; d++)
					cout << (*wIter)[d] << " ";
				cout <<endl;
				#endif
			}
			else if (dominates((*wIter), (*pIter)))
			{
				#ifdef DEBUG
				cout << "case 3\n";
				#endif

				check = false;
				break;
			}
			else
				++wIter;
		}

		if (check)
		{
			if (window.size() < wsize)
			{
				#ifdef DEBUG
				cout << "case 4\n";
				#endif

				window.push_back((*pIter));
			}
			else
			{
				#ifdef DEBUG
				cout << "case 5\n";
				#endif

				tempFile.push_back((*pIter));
				countOut++;
			}
		}
		points.erase(pIter);

		#ifdef DEBUG
		//getchar();
		cout << "window: " << endl;
		//printPoints(window);
		for (int p = 0; p < window.size(); p++)
		{
			cout << "point " << p << ": ";
			for (int d = 1; d < numDimensions+1; d++)
				cout << window[p][d] << " ";
			cout << " time stamp: " << timestamp.find(window[p][0])->second << endl;
		}
		cout << endl;
		cout << "temp file: " << endl;
		//printPoints(tempFile);
		for (int p = 0; p < tempFile.size(); p++)
		{
			cout << "point " << p << ": ";
			for (int d = 1; d < numDimensions+1; d++)
				cout << tempFile[p][d] << " ";
			cout << " time stamp: " << timestamp.find(tempFile[p][0])->second << endl;
		}
		cout << endl;
		#endif

		if (pIter == points.end())
		{
			points = vector<int*>(tempFile);
			prevCountIn = points.size();
			countIn = countOut = 0;
			tempFile.clear();
		}
	}

	for (wIter = window.begin(); wIter != window.end(); ++wIter)
		skyline.push_back((*wIter));

	#ifdef DEBUG
	cout << "skyline size: " << skyline.size() << endl;
	#endif


	return skyline;
}


void BlockNestedLoops::printPoints(vector<int*> points)
{
	for (int p = 0; p < points.size(); p++)
	{
		cout << "point " << p << ": ";
		for (int d = 0; d < numDimensions+1; d++)
			cout << points[p][d] << " ";
		cout << endl;
	}
}


BlockNestedLoops::~BlockNestedLoops()
{
	delete[] dimSizes;
	for (int j = 0; j < points.size(); j++)
		delete[] points[j];
}
