#include "BnL_cache.h"



BlockNestedLoopsForCachePoints::BlockNestedLoopsForCachePoints(vector<CacheBlock> cp, int* query, int numDimensions)
{
	this->numDimensions = numDimensions;
	points = vector<CacheBlock>(cp);
	this->query = query;
}


bool BlockNestedLoopsForCachePoints::dominates(int *a, int *b)
{
	bool check = false;

	#ifdef DEBUG
	cout << "does ";
	for (int d = 0; d < numDimensions; d++)
		cout << a[d] << " ";
	cout << " dominates ";
	for (int d = 0; d < numDimensions; d++)
		cout << b[d] << " ";
	#endif

	for (int d = 0; d < numDimensions; d++)
	{
		int aq = abs(a[d]-query[d]);
		int bq = abs(b[d]-query[d]);
		if (aq > bq)
		{
			#ifdef DEBUG
			cout << " NO";
			cout <<  endl;
			#endif

			return false;
		}
		else if (aq < bq)
			check = true;
	}

	#ifdef DEBUG
	if (check)
		cout << " YES";
	else
		cout << " NO";
	cout <<  endl;
	#endif
	
	return check;
}


bool BlockNestedLoopsForCachePoints::isDominatedBy(int *a, int *b)
{
	bool check = false;
	for (int d = 0; d < numDimensions; d++)
	{
		int aq = abs(a[d]-query[d]);
		int bq = abs(b[d]-query[d]);
		if (aq < bq)
			return false;
		else if (aq > bq)
			check = true;
	}
	
	return check;
}


vector<CacheBlock> BlockNestedLoopsForCachePoints::getSkyline(int wsize)
{
	#ifdef PRINT_RESULTS
	cout << endl << "Skyline block nested loops method" << endl;
	#endif

	vector<CacheBlock>::iterator pIter;
	vector<CacheBlock> skyline;
	vector<CacheBlock> window;
	vector<CacheBlock>::iterator wIter;
	vector<CacheBlock> tempFile;
	int countIn = 0, countOut = 0;
	int prevCountIn = 0;
	map<int, int> timestamp;
	map<int, int>::iterator sIter;

	pIter = points.begin();
	while (pIter != points.end())
	{
		#ifdef DEBUG
		cout << "examine point: ";
		for (int d = 0; d < numDimensions; d++)
			cout << pIter->query[d] << " ";
		cout << " time stamp: " << countOut << " countIn: " << countIn << endl;
		#endif

		wIter = window.begin();
		while (wIter != window.end())
		{
			sIter = timestamp.find(wIter->id);
			if (sIter->second == countIn)
			{
				#ifdef DEBUG
				cout << "case 1\n";
				cout << "output point: ";
				for (int d = 0; d < numDimensions; d++)
					cout << wIter->query[d] << " ";
				cout << " time stamp: " <<  countIn << endl;
				#endif

				skyline.push_back((*wIter));
				window.erase(wIter);
			}

			// Time stamp correction.
			else if ((sIter->second >= points.size()) && (countIn == 0))
			{
				sIter->second = prevCountIn-1;
				++wIter;
			}
			else
				++wIter;
		}
		
		sIter = timestamp.find(pIter->id);
		if (sIter == timestamp.end())
			timestamp.insert(make_pair(pIter->id, countOut));
		else
			sIter->second = countOut;
		countIn++;
		bool check = true;

		wIter = window.begin();
		while (wIter != window.end())
		{
			if (dominates(pIter->query, wIter->query))
			{
				#ifdef DEBUG
				cout << "case 2\n";
				#endif

				window.erase(wIter);

				#ifdef DEBUG
				cout << "erase point: ";
				for (int d = 0; d < numDimensions; d++)
					cout << wIter->query[d] << " ";
				cout <<endl;
				#endif
			}
			else if (dominates(wIter->query, pIter->query))
			{
				#ifdef DEBUG
				cout << "case 3\n";
				#endif

				check = false;
				break;
			}
			else
				++wIter;
		}

		if (check)
		{
			if (window.size() < wsize)
			{
				#ifdef DEBUG
				cout << "case 4\n";
				#endif

				window.push_back((*pIter));
			}
			else
			{
				#ifdef DEBUG
				cout << "case 5\n";
				#endif

				tempFile.push_back((*pIter));
				countOut++;
			}
		}
		points.erase(pIter);

		#ifdef DEBUG
		//getchar();
		cout << "window: " << endl;
		//printPoints(window);
		for (int p = 0; p < window.size(); p++)
		{
			cout << "point " << p << ": ";
			for (int d = 0; d < numDimensions; d++)
				cout << window[p].query[d] << " ";
			cout << " time stamp: " << timestamp.find(window[p]->id)->second << endl;
		}
		cout << endl;
		cout << "temp file: " << endl;
		//printPoints(tempFile);
		for (int p = 0; p < tempFile.size(); p++)
		{
			cout << "point " << p << ": ";
			for (int d = 0; d < numDimensions; d++)
				cout << tempFile[p].query[d] << " ";
			cout << " time stamp: " << timestamp.find(tempFile[p]->id)->second << endl;
		}
		cout << endl;
		#endif

		if (pIter == points.end())
		{
			points = vector<CacheBlock>(tempFile);
			prevCountIn = points.size();
			countIn = countOut = 0;
			tempFile.clear();
		}
	}

	for (wIter = window.begin(); wIter != window.end(); ++wIter)
		skyline.push_back((*wIter));

	#ifdef DEBUG
	cout << "skyline size: " << skyline.size() << endl;
	#endif


	return skyline;
}


void BlockNestedLoopsForCachePoints::printPoints(vector<CacheBlock> points)
{
	for (int p = 0; p < points.size(); p++)
	{
		cout << "point " << p << ": ";
		for (int d = 0; d < numDimensions; d++)
			cout << points[p].query[d] << " ";
		cout << endl;
	}
}


BlockNestedLoopsForCachePoints::~BlockNestedLoopsForCachePoints()
{
	for (int j = 0; j < points.size(); j++)
	{
		delete[] points[j].query;
		delete[] points[j].omasks;
	}
}
