#include "DnC.h"



// Global
int dim, median;

bool comparePoints(int* a, int* b)
{
	//#ifdef DEBUG
	//cout << "compare " << a[dim] << ", " << b[dim] << endl;
	//#endif
	
	return (a[dim] < b[dim]);
}


bool compareMedian(int *a)
{
	return (a[dim] < median);
}




DivideAndConquer::DivideAndConquer()
{
}


DivideAndConquer::DivideAndConquer(char *infile)
{
	fstream fin;
	fin.open(infile, ios::in);
	fin >> numDimensions; // read dimensionality

	// read dimension sizes
	dimSizes = new int[numDimensions];
	for (int d = 0; d < numDimensions; d++)
	{ 
		fin >> dimSizes[d];
		dimSizes[d]++;

		#ifdef DEBUG
		cout << dimSizes[d] << " ";
		#endif
	}
	#ifdef DEBUG
	cout << endl;
	#endif

	// read number of points
	fin >> numPoints; 

	// read points
	int* point;
	for (int p = 0; p < numPoints; p++)
	{
		point = new int[numDimensions];

		#ifdef DEBUG
		cout << "point " << p << ": ";
		#endif

		for (int d = 0; d < numDimensions; d++)
		{
			fin >> point[d];

			#ifdef DEBUG
			cout << point[d] << " ";
			#endif
		}
		points.push_back(point);

		#ifdef DEBUG
		cout << endl;
		#endif
	}
	fin.close();

	#ifdef DEBUG
	cout << "numDimensions=" << numDimensions << " numPoints=" << numPoints << endl;
	#endif

}


deque<int*> DivideAndConquer::getSkylineBasic(int dimension)
{
	#ifdef PRINT_RESULTS
	cout << endl << "Skyline basic divide and conquer method" << endl;
	#endif

	return getSkylineBasic(points, dimension);
}


deque<int*> DivideAndConquer::getSkylineBasic(deque<int*> points, int dimension)
{
	deque<int*> skyline;
	deque<int*>::iterator pIter;

	if (points.size() == 1)
		return points;
	else if (points.size() == 0)
		return skyline;

	//printPoints(points);
	dim = dimension;
	int pivotIndex = getMedian(points);
	nth_element(points.begin(), points.begin()+pivotIndex, points.end(), comparePoints);
	median = points[pivotIndex][dimension];

	/*cout << "medianIndex: " << pivotIndex << " median: ";
	for (int d = 0; d < numDimensions; d++)
		cout << points[pivotIndex][d] << " ";
	cout << endl;*/
	#ifdef DEBUG
	cout << endl << "median in dimension " << dimension << ": " << median << endl;
	#endif

	pIter = partition(points.begin(), points.end(), compareMedian);

	// In case points have equal values in dimension, then we iteratively perform partitioning over the previous dimensions.
	while ((pIter == points.begin()) && (dim >= 0))
	{
	//	pIter += pivotIndex;
		dim--;
		nth_element(points.begin(), points.begin()+pivotIndex, points.end(), comparePoints);
		median = points[pivotIndex][dim];
		
		#ifdef DEBUG
		cout << endl << "median in dimension " << dim << ": " << median << endl;
		#endif

		pIter = partition(points.begin(), points.end(), compareMedian);
	}

	// Points have equal values in all previous dimensions. Thus, we split points in the middle.
	if (dim < 0)
		pIter = points.begin()+pivotIndex;


	deque<int*> P1(points.begin(),	pIter);
	deque<int*> P2(pIter, 		points.end());

	#ifdef DEBUG
	cout << "Partition 1: " << endl;
	printPoints(P1);
//	cout << "Partition 2: " << endl;
//	printPoints(P2);
	#endif

	skyline = getSkylineBasic(P1, dimension);

	#ifdef DEBUG
	cout << "Partition 2: " << endl;
	printPoints(P2);
	#endif

	deque<int*> S2 = getSkylineBasic(P2, dimension);
	deque<int*> R = mergeBasic(skyline, S2, dimension);

	skyline.insert(skyline.end(), R.begin(), R.end());
	R.clear();
	S2.clear();
	
	#ifdef DEBUG
	cout << "Skyline: " << endl;
	printPoints(skyline);
	cout << endl;
//	getchar();
	#endif


	return skyline;
}


deque<int*> DivideAndConquer::mergeBasic(deque<int*> S1, deque<int*> S2, int dimension)
{
	#ifdef DEBUG
	cout << "merge" << endl;
	printPoints(S1);
	cout << "with" << endl;
	printPoints(S2);
	cout << " over dimension " << dimension << endl;
	#endif

	deque<int*> mskyline;

	if (S1.size() == 0)
	{
		#ifdef DEBUG
		cout << "merge result: " << endl;
		printPoints(S2);
		cout << endl;
		#endif

		return S2;
	}
	else if (S2.size() == 0)
	{
		#ifdef DEBUG
		cout << "merge result: " << endl;
		printPoints(S2);
		cout << endl;
		#endif

		return S2;
	}
	else if (S1.size() == 1)
	{
		#ifdef DEBUG
		cout << "\t\t\tcase 1\n";
		#endif

		for (int p = 0; p < S2.size(); p++)
		{
			if (!dominates(S1[0], S2[p]))
				mskyline.push_back(S2[p]);

			#ifdef DEBUG
			else
			{
				cout << "\t\t\t\tdrop point: ";
				for (int d = 0; d < numDimensions; d++)
					cout << S2[p][d] << " ";
				cout << endl;
			}
			#endif
		}
	}
	else if (S2.size() == 1)
	{
		#ifdef DEBUG
		cout << "\t\t\tcase 2\n";
		#endif

		for (int p = 0; p < S1.size(); p++)
		{
			if (dominates(S1[p], S2[0]))
			{
				#ifdef DEBUG
				cout << "merge result: " << endl;
				printPoints(mskyline);
				cout << endl;
				#endif

				return mskyline;
			}
		}

		mskyline.push_back(S2[0]);
	}
	else if (dimension == 1)
	{
		#ifdef DEBUG
		cout << "\t\t\tcase 3\n";
		#endif

		dim = 0;
		deque<int*>::iterator min = min_element(S1.begin(), S1.end(), comparePoints);
		for (int p = 0; p < S2.size(); p++)
		{
			if (S2[p][dim] < (*min)[dim])
				mskyline.push_back(S2[p]);
		}
	}
	else
	{
		#ifdef DEBUG
		cout << "\t\t\tcase 4\n";
		#endif

		int pivotIndex = getMedian(S1);
		dim = dimension-1;
		nth_element(S1.begin(), S1.begin()+pivotIndex, S1.end(), comparePoints);
		median = S1[pivotIndex][dim];

		#ifdef DEBUG
		cout << endl << "median in dimension " << dim << ": " << median << endl;
		#endif

		deque<int*>::iterator sIter = partition(S1.begin(), S1.end(), compareMedian);

		// In case points have equal values in dimension, then we iteratively perform partitioning over the previous dimensions.
		while ((sIter == S1.begin()) && (dim >= 0))
		{
			dim--;
			nth_element(S1.begin(), S1.begin()+pivotIndex, S1.end(), comparePoints);
			median = S1[pivotIndex][dim];
		
			#ifdef DEBUG
			cout << endl << "median in dimension " << dim << ": " << median << endl;
			#endif

			sIter = partition(S1.begin(), S1.end(), compareMedian);
		}

		// Points have equal values in all previous dimensions. Thus, we split points in the middle.
		if (dim < 0)
			sIter = S1.begin()+pivotIndex;

		//if (sIter == S1.begin())
		//	sIter += pivotIndex;

		
		deque<int*> S11(S1.begin(), 	sIter);
		deque<int*> S12(sIter,		S1.end());

		pivotIndex = getMedian(S2);
		sIter = partition(S2.begin(), S2.end(), compareMedian);
		//if (sIter == S2.begin())
		//	sIter = S2.begin()+pivotIndex;
		deque<int*> S21(S2.begin(), 	sIter);
		deque<int*> S22(sIter, 		S2.end());

		#ifdef DEBUG
		cout << "S11:" << endl;
		printPoints(S11);
		cout << "S12:" << endl;
		printPoints(S12);
		cout << "S21:" << endl;
		printPoints(S21);
		cout << "S22:" << endl;
		printPoints(S22);
		#endif

		deque<int*> R1 = mergeBasic(S11, S21, dimension);
		deque<int*> R2 = mergeBasic(S12, S22, dimension);
		deque<int*> R3 = mergeBasic(S11, R2, dimension-1);
		mskyline.insert(mskyline.end(), R1.begin(), R1.end());
		mskyline.insert(mskyline.end(), R3.begin(), R3.end());
		R1.clear();
		R2.clear();
		R3.clear();
	}

	#ifdef DEBUG
	cout << "merge result: " << endl;
	printPoints(mskyline);
	cout << endl;
	#endif


	return mskyline;
}


int DivideAndConquer::getMedian(deque<int*> points)
{
	if ((points.size()%2) == 1) 
		return iround((double)points.size()/2)-1;
	return iround((double)points.size()/2);
}


bool DivideAndConquer::dominates(int *a, int *b)
{
	bool check = false;

	#ifdef DEBUG
	cout << "\t\t\t\tdoes ";
	for (int d = 0; d < numDimensions; d++)
		cout << a[d] << " ";
	cout << " dominates ";
	for (int d = 0; d < numDimensions; d++)
		cout << b[d] << " ";
	#endif

	for (int d = 0; d < numDimensions; d++)
	{
		if (a[d] > b[d])
		{
			#ifdef DEBUG
			cout << " NO";
			cout <<  endl;
			#endif

			return false;
		}
		else if (a[d] < b[d])
			check = true;
	}

	#ifdef DEBUG
	if (check)
		cout << " YES";
	else
		cout << " NO";
	cout <<  endl;
	#endif
	
	return check;
}


void DivideAndConquer::printPoints(deque<int*> points)
{
	for (int p = 0; p < points.size(); p++)
	{
		cout << "point " << p << ": ";
		for (int d = 0; d < numDimensions; d++)
			cout << points[p][d] << " ";
		cout << endl;
	}
}


DivideAndConquer::~DivideAndConquer()
{
	delete[] dimSizes;
	for (int j = 0; j < points.size(); j++)
		delete[] points[j];
}
