#include "dynamic_skyline.h"
#include "timing.h"
//#define DEBUG



bool queryCacheComparator(const CacheBlock &lhs, const CacheBlock &rhs)
{
	return lhs.comparisonKey > rhs.comparisonKey;
}




DynamicSkyline::DynamicSkyline(int numDimensions, int *dimSizes, int numPoints, int** points, int qcache_size, int cpolicy)
{
	init(numDimensions, dimSizes, numPoints, points, qcache_size, cpolicy);
}


void DynamicSkyline::init(int numDimensions, int *dimSizes, int numPoints, int** points, int qcache_size, int cpolicy)
{
	this->MAX_CACHE_SIZE = qcache_size;
	this->numDimensions = numDimensions;
	this->numPoints = numPoints;
	this->dimSizes = dimSizes;
	this->points = points;
	this->cpolicy = cpolicy;
	
	sumDimSizes = new int[numDimensions];
	sumDimSizes[0] = 0;
	for (int i = 1; i < numDimensions; i++)
	{
		sumDimSizes[i] = sumDimSizes[i-1] + dimSizes[i-1];
	}
	
	length = sumDimSizes[numDimensions-1] + dimSizes[numDimensions-1];
	numOMasks = (int)pow(2.0, (double)numDimensions);
	
	bslices = new boost::dynamic_bitset<>[length];
	omasks = new boost::dynamic_bitset<>[numOMasks];
	mask = dbitset(numPoints);
	containedIn = new int[numPoints];
	osizes = new int[numOMasks];
	oskyline_sizes = new int[numOMasks];
	if (cpolicy != FIFO)
		make_heap(cache.begin(), cache.end(), queryCacheComparator);

	/*for (int i=0; i<length; i++)
	{
		bslices[i] = dbitset(numPoints, 0ul);
	}
	
	for (int i=0; i<numOMasks; i++)
	{
		omasks[i] = dbitset(numPoints, 0ul);
	}

	mask.set();
	dbitset temp(numDimensions);
	for (int p = 0; p < numPoints; p++)
	{
		temp.reset();
		for (int d = 0; d < numDimensions; d++)
		{
			int val = points[p][d]-query[d];
			for (int k = abs(val); k < dimSizes[d]; k++)
			//for (int k = points[p][d]; k < dimSizes[d]; k++)
				bslices[sumDimSizes[d]+k][p] = true;

			if (val < 0) 
				temp[numDimensions-d-1] = true;		// numDimensions-d-1 denotes xy dimension order and not yx in orthants ids
			points[p][d] = abs(val);
		}
		containedIn[p] = temp.to_ulong();
		omasks[containedIn[p]][p] = true;
	}*/
}


DynamicSkyline::DynamicSkyline(char *infile, int qcache_size, int cpolicy)
{
	fstream fin;
	fin.open(infile, ios::in);
	fin >> numDimensions; // read dimensionality
	dimSizes = new int[numDimensions];
	for (int d=0; d<numDimensions; d++){ // read dimension sizes
		fin >> dimSizes[d];
		dimSizes[d]++;
//		cout << dimSizes[d] << " ";
	}
//	cout << endl;

	fin >> numPoints; // read number of points

	// read points
	points = new int*[numPoints];
	for (int p = 0; p < numPoints; p++)
	{
		points[p] = new int[numDimensions];

		#ifdef DEBUG
		cout << "point " << p << ": ";
		#endif

		for (int d = 0; d < numDimensions; d++)
		{
			fin >> points[p][d];

			#ifdef DEBUG
			cout << points[p][d] << " ";
			#endif
		}

		#ifdef DEBUG
		cout << endl;
		#endif
	}
	fin.close();

	#ifdef DEBUG
	cout << "numDimensions=" << numDimensions << " numPoints=" << numPoints << endl;
	#endif

	init(numDimensions, dimSizes, numPoints, points, qcache_size, cpolicy);
}


dbitset DynamicSkyline::isDominated(dbitset a, dbitset b)
{
	return a&b;
}


dbitset DynamicSkyline::dominates(dbitset a, dbitset b)
{
	return ~(a|b);
}


void DynamicSkyline::getA_B(int *point, dbitset *a, dbitset *b, int *query)
{
	a->set();
	b->reset();
	for (int d=0; d<numDimensions; d++)
	{
		int val = abs(point[d]-query[d]);
		//int val = point[d];
		(*a) &= bslices[sumDimSizes[d]+val];
		if (val!=0)
		{
			(*b) |= bslices[sumDimSizes[d]+val-1];
		}
	}
}


dbitset* DynamicSkyline::getDynamicOrthantsSkylineBitmap(int *query)
{
	dbitset* R = new boost::dynamic_bitset<>[numOMasks];
	for (int i = 0; i < numOMasks; i++)
		R[i] = dbitset(numPoints, 0ul);

	for (int p = 0; p < numPoints; p++)
	{
		int orthantID = containedIn[p];
		if (!omasks[orthantID][p])
		{
			mask.reset(p);
			pointsPruned++;			
			continue;
		}

		/*dbitset a(numPoints), b(numPoints);
		a = mask;
		getA_B(points[p], &a, &b, query);
		dbitset c = isDominated(a, b);*/
		dbitset c = isDominated(points[p], query);

		// Uncomment the following line to activate reverse dominance test
		//dbitset d = dominates(a, b);

		if ((c&omasks[orthantID]).any())
		{
			mask.reset(p);
			omasks[orthantID].reset(p);
			R[orthantID].set(p);
		}
		else if (c.any())
		{
			mask.reset(p);
			omasks[orthantID].reset(p);
			R[orthantID].set(p);
			oskyline_sizes[orthantID]++;
		}
		else
		{
			mask.set(p);
			oskyline_sizes[orthantID]++;
		}

		// Uncomment the following two lines to activate reverse dominance test
		//mask &= ~d;
		//omasks[orthantID] &= ~d;
	}

	return R;
}


void DynamicSkyline::calculateMasksDeMorgan(int *query)
{
	memset(containedIn, 0, numPoints);
	for (int i=0; i<length; i++)
		bslices[i] = dbitset(numPoints, 0ul);

	for (int i=0; i<numOMasks; i++)
	{
		omasks[i] = dbitset(numPoints, 0ul);
		osizes[i] = 0;
		oskyline_sizes[i] = 0;
	}

	// Update bitmap table and orthant masks
	dbitset temp(numDimensions);
	for (int p = 0; p < numPoints; p++)
	{
		temp.reset();
		for (int d = 0; d < numDimensions; d++)
		{
			int val = points[p][d]-query[d];
			for (int k = abs(val); k < dimSizes[d]; k++)
				bslices[sumDimSizes[d]+k][p] = true;
			if (val < 0)
				temp[numDimensions-d-1] = true;
		}
		containedIn[p] = (int)temp.to_ulong();
		//mask.set(p);
		omasks[containedIn[p]][p] = true;
		osizes[containedIn[p]]++;
	}

	// Exploit query cache
	if ((cpolicy == LPP) || (cpolicy == FIFO))
	{
		for (int q = 0; q < cache.size(); q++)
		{
			int orthant = inWhichOrthantWRT(cache[q].query, query);
			//dbitset notO = ~cache[q].omasks[orthant];
			//mask &= notO;
			omasks[orthant] &= ~cache[q].omasks[orthant];
		}
	}
	else
	{
		BlockNestedLoopsForCachePoints *bnlc = new BlockNestedLoopsForCachePoints(cache, query, numDimensions);
		vector<CacheBlock> cskyline = bnlc->getSkyline(cache.size());

		#ifdef DEBUG
		cout << "cache:" << endl;
		for (int j = 0; j < cache.size(); j++)
		{
			for (int d = 0; d < numDimensions; d++)
				cout << cache[j].query[d] << " ";
			cout << "comparison key: " << cache[j].comparisonKey << endl;
		}
		cout << endl;

		cout << "cache skyline:" << endl;
		for (int j = 0; j < cskyline.size(); j++)
		{
			for (int d = 0; d < numDimensions; d++)
				cout << cskyline[j].query[d] << " ";
			cout << "comparison key: " << cskyline[j].comparisonKey << endl;
		}
		cout << endl;
		#endif

		for (int q = 0; q < cskyline.size(); q++)
		{
			int orthant = inWhichOrthantWRT(cache[cskyline[q].id].query, query);
			omasks[orthant] &= ~cache[cskyline[q].id].omasks[orthant];

			if (cpolicy == LRU)
				cache[cskyline[q].id].comparisonKey = numQueries;
			else
				cache[cskyline[q].id].comparisonKey++;
		}
		//delete bnl;
	}
}


void DynamicSkyline::calculateMasks(int *query)
{
	//dbitset R(numPoints, 0ul);
	dbitset* oR = new boost::dynamic_bitset<>[numOMasks];

	memset(containedIn, 0, numPoints);
	for (int i=0; i<length; i++)
		bslices[i] = dbitset(numPoints, 0ul);

	for (int i=0; i<numOMasks; i++)
	{
		omasks[i] = dbitset(numPoints, 0ul);
		oR[i] = dbitset(numPoints, 0ul);
		osizes[i] = 0;
		oskyline_sizes[i] = 0;
	}

	// Update bitmap table and orthant masks
	dbitset temp(numDimensions);
	for (int p = 0; p < numPoints; p++)
	{
		temp.reset();
		for (int d = 0; d < numDimensions; d++)
		{
			int val = points[p][d]-query[d];
			for (int k = abs(val); k < dimSizes[d]; k++)
				bslices[sumDimSizes[d]+k][p] = true;
			if (val < 0)
				temp[numDimensions-d-1] = true;
		}
		containedIn[p] = (int)temp.to_ulong();
		//mask.set(p);
		omasks[containedIn[p]][p] = true;
		osizes[containedIn[p]]++;
	}

	// Exploit query cache
	if ((cpolicy == LPP) || (cpolicy == FIFO))
	{
		for (int q = 0; q < cache.size(); q++)
		{
			int orthant = inWhichOrthantWRT(cache[q].query, query);
			//dbitset notO = ~cache[q].omasks[orthant];
			//mask &= notO;
			omasks[orthant] &= ~cache[q].omasks[orthant];
		}
	}
	else
	{
		BlockNestedLoopsForCachePoints *bnlc = new BlockNestedLoopsForCachePoints(cache, query, numDimensions);
		vector<CacheBlock> cskyline = bnlc->getSkyline(cache.size());

		#ifdef DEBUG
		cout << "cache:" << endl;
		for (int j = 0; j < cache.size(); j++)
		{
			for (int d = 0; d < numDimensions; d++)
				cout << cache[j].query[d] << " ";
			cout << "comparison key: " << cache[j].comparisonKey << endl;
		}
		cout << endl;

		cout << "cache skyline:" << endl;
		for (int j = 0; j < cskyline.size(); j++)
		{
			for (int d = 0; d < numDimensions; d++)
				cout << cskyline[j].query[d] << " ";
			cout << "comparison key: " << cskyline[j].comparisonKey << endl;
		}
		cout << endl;
		#endif

		for (int q = 0; q < cskyline.size(); q++)
		{
			int orthant = inWhichOrthantWRT(cache[cskyline[q].id].query, query);
			//R |= cache[ckyline[q].id].omasks[orthant];
			oR[orthant] |= cache[cskyline[q].id].omasks[orthant];

			if (cpolicy == LRU)
				cache[cskyline[q].id].comparisonKey = numQueries;
			else
				cache[cskyline[q].id].comparisonKey++;
		}
		//delete bnl;
	}

	if (cache.size() > 0)
	{
		//mask &= ~R;
		for (int i=0; i<numOMasks; i++)
			omasks[i] &= ~oR[i];
	}

	delete[] oR;
}


/*void DynamicSkyline::calculateMasksMap(int* query)
{
	memset(containedIn, 0, numPoints);

	for (int i=0; i<length; i++)
		bslices[i] = dbitset(numPoints, 0ul);

	for (int i=0; i<numOMasks; i++)
	{
		omasks[i] = dbitset(numPoints, 0ul);
		osizes[i] = 0;
		oskyline_sizes[i] = 0;
	}

	// orthant qi resides in w.r.t. to q, i.e. {Qj} set
	multimap<int, int> cachePartition;
	multimap<int, int>::iterator cpIter;
	for (int q = 0; q < cache.size(); q++)
	{
		int key = inWhichOrthantWRT(cache[q].query, query);
		cachePartition.insert(make_pair(key, q));
	}

	dbitset temp(numDimensions);
	for (int p = 0; p < numPoints; p++)
	{
		temp.reset();
		for (int d = 0; d < numDimensions; d++)
		{
			int val = points[p][d]-query[d];
			for (int k = abs(val); k < dimSizes[d]; k++)
				bslices[sumDimSizes[d]+k][p] = true;
			if (val < 0)
				temp[numDimensions-d-1] = true;
		}
		containedIn[p] = (int)temp.to_ulong();
		//mask.set(p);
		omasks[containedIn[p]][p] = true;
		osizes[containedIn[p]]++;

		if (cache.size() > 0)
		{
			int numQueries = 0;
			cpIter = cachePartition.lower_bound(containedIn[p]);
			while ((cpIter != cachePartition.end()) && (cpIter->first == containedIn[p]))
			{
				int orth = inWhichOrthantWRT(points[p], cache[cpIter->second].query);
				if (orth == containedIn[p])
				{
					if (cpolicy == LRU)
						cache[cpIter->second].comparisonKey = numQueries;
					else if (cpolicy == LFU)
						cache[cpIter->second].comparisonKey++;

					// Cache pruning
					if (!cache[cpIter->second].omasks[orth][p])
					{
						//mask.reset(p);
						omasks[containedIn[p]].reset(p);
						break;
					}
				}
				++cpIter;
			}
		}
	}
}*/


dbitset DynamicSkyline::getDynamicSkylineCachingBitmap(int* query)
{
	#ifdef PRINT_RESULTS
	if (toCacheOrNotToCache)
		cout << endl << "Dynamic skyline via caching bitmap method" << endl;
	else
		cout << endl << "Dynamic skyline bitmap method" << endl;
	#endif

	pointsPruned = 0;
	numQueries++;

	// Initialize masks and construct bitmap table
	Timing *tim = new Timing();
	tim->startClock();
	calculateMasksDeMorgan(query);
	cout << "calculate masks time: " << tim->stopClock() << endl;

	// Calculate dynamic and orthant skylines
	tim->startClock();
	dbitset* O = getDynamicOrthantsSkylineBitmap(query);
	cout << "calculate dynamic and orthant skylines time: " << tim->stopClock() << endl;


	// Update cache
	tim->startClock();
	if (cpolicy != NO_CACHE)
	{
		if (cache.size() >= MAX_CACHE_SIZE)
		{
			if (cpolicy == FIFO)
				cache.erase(cache.begin());
			else
			{
				pop_heap(cache.begin(), cache.end(), queryCacheComparator);
				cache.pop_back();
			}
		}
		CacheBlock c(cache.size(), query, O, numDimensions, numOMasks, numPoints);

		if (cpolicy == LPP)
		{
			int pp = 0;
			for (int o = 0; o < numOMasks; o++)
			{
				dbitset b(numDimensions, (unsigned long)o);
				pp += osizes[(b.flip()).to_ulong()]*oskyline_sizes[o];
			}
			c.comparisonKey = pp;
		}

		cache.push_back(c);
		//if (cpolicy != FIFO)
		push_heap(cache.begin(), cache.end(), queryCacheComparator);

		//cout << endl << printCachePoints();
	}
	cout << "update cache time: " << tim->stopClock() << endl;

	delete tim;


	return mask;
}


int DynamicSkyline::inWhichOrthantWRT(int* p1, int* p2)
{
	dbitset temp(numDimensions, 0ul);
	for (int d = 0; d < numDimensions; d++)
	{
		if (p1[d]-p2[d] < 0) 
			temp[numDimensions-d-1] = true;		// numDimensions-d-1 denotes xy dimension order and not yx in orthants ids

	}

	return (int)(temp.to_ulong());
}


dbitset DynamicSkyline::isDominated(int *point, int* query)
{
	//dbitset a = mask;
	dbitset a(numPoints);
	a.set();
	dbitset b(numPoints, 0ul);
	//b.reset();
	for (int d=0; d<numDimensions; d++){
		int val = abs(point[d]-query[d]);
		a &= bslices[sumDimSizes[d]+val];
		if (val!=0){
			b |= bslices[sumDimSizes[d]+val-1];
		}
	}
	return a&b;
}


dbitset DynamicSkyline::getDynamicSkylineBitmap(int* query)
{
	for (int p=0; p<numPoints; p++)
	{
		if (!mask[p])	{
			continue;
		}
		dbitset isdom = isDominated(points[p], query);
		if (isdom.any()) { // not in the skyline
			mask.reset(p);
		}
	}
	return mask;
}


string DynamicSkyline::toString()
{
	ostringstream sout;
	for (int i = 0; i < length; i++)
		sout << "slice " << i << ": " << printSlice(i) << endl;

	for (int i = 0; i < numOMasks; i++)
		cout << "orthant mask " << i << ": " << printOrthantMask(i) << endl;

	return sout.str();
}


string DynamicSkyline::printSlice(int slice)
{
	ostringstream sout;
	for (int p=0; p<numPoints; p++)
	{
		sout << bslices[slice][p];
	}
	return sout.str();
}


string DynamicSkyline::printOrthantMask(int mask)
{
	ostringstream sout;
	for (int p=0; p<numPoints; p++)
	{
		sout << omasks[mask][p];
	}
	return sout.str();
}


string DynamicSkyline::printCachePoints()
{
	ostringstream sout;
	sout << "Query cache";

	switch (cpolicy)
	{
	case LRU:
		sout << " (LRU policy)";
		break;
	case LFU:
		sout << " (LFU policy)";
		break;
	case LPP:
		sout << " (LPP policy)";
		break;
	case FIFO:
		sout << " (FIFO policy)";
		break;
	}
	sout << endl;

	for (int q = 0; q < cache.size(); q++)
	{
		sout << "\tquery point: ";
		for (int d = 0; d < numDimensions; d++)
			sout << cache[q].query[d] << " ";
		sout << "comparison key: " << cache[q].comparisonKey << endl;
	}
	return sout.str();
}


DynamicSkyline::~DynamicSkyline()
{
	delete[] dimSizes;
	delete[] sumDimSizes;
	delete[] bslices;
	delete[] containedIn;
	delete[] osizes;
	delete[] oskyline_sizes;
	for (int j = 0; j < numPoints; j++)
		delete[] points[j];
	delete[] omasks;
	for (int j = 0; j < cache.size(); j++)
	{
		delete[] cache[j].query;
		delete[] cache[j].omasks;
	}
}
