#ifndef _DYNAMIC_SKYLINES_H_
#define _DYNAMIC_SKYLINES_H_

#include "def.h"
#include <sstream>
#include "math.h"
#include <map>
#include "BnL_cache.h"

using namespace boost;
using namespace std;



/*class PriorityQueueComparator
{
public:
	bool operator() (const CacheBlock &lhs, const CacheBlock &rhs) const
	{
		return lhs.comparisonKey > rhs.comparisonKey;
	}
};*/




class DynamicSkyline
{
//	**** important for dynamic_bitset ****
//	Note that operator[] goes from the least-significant bit at 0 to the most significant bit at size()-1.
//	The operator<< for dynamic_bitset prints the bitset from most-significant to least-significant,
//	since that is the format most people are use to reading.
//	printSlice, defined here, prints from least-significant to most-significant
public:
	int numDimensions; // number of dimensions
	int *dimSizes; // dimension sizes
	int *sumDimSizes; // prefix sum array of dimension sizes, sumDimSizes[0]=0, sumDimSizes[1]=dimSizes[0], etc.
	int length; // number of bitslices, sum of dimSizes
	int numOMasks; // numebr of orthant masks
	int numPoints; // bitslice size
	int **points; // the points
	dbitset mask; // reverse dominace mask
	dbitset *bslices; // array of bitslices
	dbitset *omasks; // array of orthant masks
	int *containedIn; // array indicating the orthant where a point is contained

	int pointsPruned;

	int MAX_CACHE_SIZE;
	vector<CacheBlock> cache; // query cache
	// For cache replacement policies
	int cpolicy;
	int numQueries;
	int* osizes;
	int* oskyline_sizes;
	
	DynamicSkyline(int numDimensions, int *dimSizes, int numPoints, int** points, int qcache_size, int cpolicy);
	DynamicSkyline(char *infile, int qcache_size, int cpolicy);
	void init(int numDimensions, int *dimSizes, int numPoints, int** points, int qcache_size, int cpolicy);

	dbitset isDominated(int *point, int *query);
	dbitset dominates(int *point);
	dbitset getDynamicSkylineBitmap(int* query);

	dbitset isDominated(dbitset a, dbitset b);
	dbitset dominates(dbitset a, dbitset b);
	void getA_B(int *point, dbitset *a, dbitset *b, int *query);
	dbitset* getDynamicOrthantsSkylineBitmap(int* query);
	void calculateMasksDeMorgan(int *query);
	void calculateMasks(int *query);
	//void calculateMasksMap(int *query);
	dbitset getDynamicSkylineCachingBitmap(int* query);

	int inWhichOrthantWRT(int* p1, int* p2); // returns the orthant where p1 resides w.r.t. p2
	
	string toString();
	string printSlice(int slice);
	string printOrthantMask(int mask);
	string printCachePoints();

	~DynamicSkyline();
};


#endif /*DYNAMIC_SKYLINES_H_*/
