#ifndef FAST_BITMAP_H_
#define FAST_BITMAP_H_

#include "def.h"
#include <boost/dynamic_bitset.hpp>
#include <sstream>

using namespace boost;
using namespace std;




typedef boost::dynamic_bitset<> dbitset;

class FastBitmap{
//	**** important for dynamic_bitset ****
//	Note that operator[] goes from the least-significant bit at 0 to the most significant bit at size()-1.
//	The operator<< for dynamic_bitset prints the bitset from most-significant to least-significant,
//	since that is the format most people are use to reading.
//	printSlice, defined here, prints from least-significant to most-significant
public:
	int numDimensions; // number of dimensions
	int *dimSizes; // dimension sizes
	int *sumDimSizes; // prefix sum array of dimension sizes, sumDimSizes[0]=0, sumDimSizes[1]=dimSizes[0], etc.
	int length; // number of bitslices, sum of dimSizes
	int numPoints; // bitslice size
	int **points; // the points
	dbitset *bslices; // array of bitslices
	
	FastBitmap(int numDimensions, int *dimSizes, int numPoints, int** points);
	FastBitmap(char *infile);
	void init(int numDimensions, int *dimSizes, int numPoints, int** points);
	
	
	dbitset isDominated(int *point);
	dbitset isDominated(dbitset a, dbitset b);
	dbitset isDominated(int *point, dbitset mask);
	dbitset dominates(dbitset a, dbitset b);
	dbitset dominates(int *point);
	void getA_B(int *point, dbitset *a, dbitset *b);
	dbitset getA(int *point, dbitset mask);
	dbitset getB(int *point);
	dbitset getSkyline(dbitset mask); // top-1 for any monotone preference function
	dbitset getSkyline();
	dbitset getSkyband(int k, dbitset mask); // top-k for any monotone preference function
	dbitset getSkyband(int k);
	dbitset getSkylayer(int l, dbitset mask); // get skyline, remove from points and repeat l-1 times 
	dbitset getSkylayer(int l);
	dbitset getSkylayers(int l, dbitset mask); // get skylayers up to l
	dbitset getSkylayers(int l);
	
	int getDominanceScore(int *point);
	
	vector<int> getIDsFromBitset(dbitset bitset);
	string printIDsFromBitset(dbitset bitset);
	
	string toString();
	string printSlice(int slice);

	~FastBitmap();
};


#endif /*FAST_BITMAP_H_*/
